@@ -12,7 +12,9 @@
 // This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 // WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 //
-//
+// $URL$
+// $Id$
+// SPDX-License-Identifier: LGPL-3.0+
 //
 // Author(s)     : Jane Tournois
 
@@ -51,7 +53,7 @@ struct Time_stamper
   static std::size_t time_stamp(const T* pt)
   {
     if(pt == NULL){
-      return 0;
+      return std::size_t(-1);
     }
     return pt->time_stamp();
   }
